package com.google.inject.internal;

import com.google.inject.spi.InjectionPoint;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;

/**
 * Produces construction proxies that invoke the class constructor.
 */
final class DefaultConstructionProxyFactory<T> implements ConstructionProxyFactory<T> {

    private final InjectionPoint injectionPoint;

    /**
     * @param injectionPoint an injection point whose member is a constructor of {@code T}.
     */
    DefaultConstructionProxyFactory(InjectionPoint injectionPoint) {
        this.injectionPoint = injectionPoint;
    }

    public ConstructionProxy<T> create() {
        @SuppressWarnings("unchecked") // the injection point is for a constructor of T
        final Constructor<T> constructor = (Constructor<T>) injectionPoint.getMember();

        // Use FastConstructor if the constructor is public.
        if (Modifier.isPublic(constructor.getModifiers())) {
            Class<T> classToConstruct = constructor.getDeclaringClass();
            if (!Modifier.isPublic(classToConstruct.getModifiers())) {
                constructor.setAccessible(true);
            }
        } else {
            constructor.setAccessible(true);
        }

        return new ConstructionProxy<T>() {
            public T newInstance(Object... arguments) throws InvocationTargetException {
                try {
                    return constructor.newInstance(arguments);
                } catch (InstantiationException e) {
                    throw new AssertionError(e); // shouldn't happen, we know this is a concrete type
                } catch (IllegalAccessException e) {
                    throw new AssertionError(e); // a security manager is blocking us, we're hosed
                }
            }

            public InjectionPoint getInjectionPoint() {
                return injectionPoint;
            }

            public Constructor<T> getConstructor() {
                return constructor;
            }
        };
    }
}
