package com.google.inject.spi;

import com.google.common.base.Objects;
import com.google.common.collect.ImmutableList;
import com.google.inject.Binder;
import com.google.inject.internal.Errors;
import com.google.inject.internal.util.SourceProvider;

import java.util.List;

import static com.google.common.base.Preconditions.checkNotNull;

/**
 * An error message and the context in which it occured. Messages are usually created internally by
 * Guice and its extensions. Messages can be created explicitly in a module using {@link
 * com.google.inject.Binder#addError(Throwable) addError()} statements:
 * <pre>
 *     try {
 *       bindPropertiesFromFile();
 *     } catch (IOException e) {
 *       addError(e);
 *     }</pre>
 *
 */
public final class Message implements Element {
    private final String message;
    private final Throwable cause;
    private final List<Object> sources;

    public Message(List<Object> sources, String message, Throwable cause) {
        this.sources = ImmutableList.copyOf(sources);
        this.message = checkNotNull(message, "message");
        this.cause = cause;
    }

    public Message(String message, Throwable cause) {
        this(ImmutableList.of(), message, cause);
    }

    public Message(Object source, String message) {
        this(ImmutableList.of(source), message, null);
    }

    public Message(String message) {
        this(ImmutableList.of(), message, null);
    }

    public String getSource() {
        return sources.isEmpty()
                ? SourceProvider.UNKNOWN_SOURCE.toString()
                : Errors.convert(sources.get(sources.size() - 1)).toString();
    }

    public List<Object> getSources() {
        return sources;
    }

    /**
     * Gets the error message text.
     */
    public String getMessage() {
        return message;
    }

    public <T> T acceptVisitor(ElementVisitor<T> visitor) {
        return visitor.visit(this);
    }

    /**
     * Returns the throwable that caused this message, or {@code null} if this
     * message was not caused by a throwable.
     */
    public Throwable getCause() {
        return cause;
    }

    @Override
    public String toString() {
        return message;
    }

    @Override
    public int hashCode() {
        return message.hashCode();
    }

    @Override
    public boolean equals(Object o) {
        if (!(o instanceof Message)) {
            return false;
        }
        Message e = (Message) o;
        return message.equals(e.message) && Objects.equal(cause, e.cause) && sources.equals(e.sources);
    }

    public void applyTo(Binder binder) {
        binder.withSource(getSource()).addError(this);
    }

}
