package com.google.inject.spi;

import com.google.common.collect.ImmutableList;
import com.google.inject.Binder;
import com.google.inject.Binding;
import com.google.inject.matcher.Matcher;

import java.util.List;

/**
 * Binds keys (picked using a Matcher) to a provision listener. Listeners are created explicitly in
 * a module using {@link Binder#bindListener(Matcher, ProvisionListener...)} statements:
 */
public final class ProvisionListenerBinding implements Element {

    private final Object source;
    private final Matcher<? super Binding<?>> bindingMatcher;
    private final List<ProvisionListener> listeners;

    ProvisionListenerBinding(Object source,
                             Matcher<? super Binding<?>> bindingMatcher,
                             ProvisionListener[] listeners) {
        this.source = source;
        this.bindingMatcher = bindingMatcher;
        this.listeners = ImmutableList.copyOf(listeners);
    }

    /**
     * Returns the registered listeners.
     */
    public List<ProvisionListener> getListeners() {
        return listeners;
    }

    /**
     * Returns the binding matcher which chooses which bindings the listener should be notified of.
     */
    public Matcher<? super Binding<?>> getBindingMatcher() {
        return bindingMatcher;
    }

    public Object getSource() {
        return source;
    }

    public <R> R acceptVisitor(ElementVisitor<R> visitor) {
        return visitor.visit(this);
    }

    public void applyTo(Binder binder) {
        binder.withSource(getSource()).bindListener(bindingMatcher,
                listeners.toArray(new ProvisionListener[listeners.size()]));
    }
}
