package com.google.inject.spi;

import com.google.inject.Binder;
import com.google.inject.TypeLiteral;
import com.google.inject.internal.Errors;
import com.google.inject.matcher.Matcher;

import static com.google.common.base.Preconditions.checkNotNull;

/**
 * Registration of type converters for matching target types. Instances are created
 * explicitly in a module using {@link com.google.inject.Binder#convertToTypes(Matcher,
 * TypeConverter) convertToTypes()} statements:
 * <pre>
 *     convertToTypes(Matchers.only(TypeLiteral.get(DateTime.class)), new DateTimeConverter());</pre>
 *
 */
public final class TypeConverterBinding implements Element {
    private final Object source;
    private final Matcher<? super TypeLiteral<?>> typeMatcher;
    private final TypeConverter typeConverter;

    public TypeConverterBinding(Object source, Matcher<? super TypeLiteral<?>> typeMatcher,
                                TypeConverter typeConverter) {
        this.source = checkNotNull(source, "source");
        this.typeMatcher = checkNotNull(typeMatcher, "typeMatcher");
        this.typeConverter = checkNotNull(typeConverter, "typeConverter");
    }

    public Object getSource() {
        return source;
    }

    public Matcher<? super TypeLiteral<?>> getTypeMatcher() {
        return typeMatcher;
    }

    public TypeConverter getTypeConverter() {
        return typeConverter;
    }

    public <T> T acceptVisitor(ElementVisitor<T> visitor) {
        return visitor.visit(this);
    }

    public void applyTo(Binder binder) {
        binder.withSource(getSource()).convertToTypes(typeMatcher, typeConverter);
    }

    @Override
    public String toString() {
        return typeConverter + " which matches " + typeMatcher
                + " (bound at " + Errors.convert(source) + ")";
    }
}
