package com.google.inject;

import static com.google.inject.internal.InternalFlags.IncludeStackTraceOption;
import static com.google.inject.internal.InternalFlags.getIncludeStackTraceOption;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertSame;
import static org.junit.Assert.assertTrue;

import com.google.common.base.Function;
import com.google.common.base.Joiner;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Iterables;
import com.google.common.testing.GcFinalization;

import java.lang.ref.ReferenceQueue;
import java.lang.ref.WeakReference;

public class Asserts {
    private Asserts() {
    }

    /**
     * Returns the String that would appear in an error message for this chain of classes
     * as modules.
     */
    public static String asModuleChain(Class... classes) {
        return Joiner.on(" -> ").appendTo(new StringBuilder(" (via modules: "),
                Iterables.transform(ImmutableList.copyOf(classes), new Function<Class, String>() {
                    @Override
                    public String apply(Class input) {
                        return input.getName();
                    }
                })).append(")").toString();
    }

    /**
     * Returns the source file appears in error messages
     */
    public static String getDeclaringSourcePart(Class clazz) {
        if (getIncludeStackTraceOption() == IncludeStackTraceOption.OFF) {
            return ".configure(Unknown Source";
        }
        return ".configure(" + clazz.getSimpleName() + ".java:";
    }

    /**
     * Returns true if {@link com.google.inject.internal.InternalFlags#getIncludeStackTraceOption()} returns {@link
     * IncludeStackTraceOption#OFF}.
     */
    public static boolean isIncludeStackTraceOff() {
        return getIncludeStackTraceOption() == IncludeStackTraceOption.OFF;
    }

    /**
     * Returns true if {@link  com.google.inject.internal.InternalFlags#getIncludeStackTraceOption()} returns {@link
     * IncludeStackTraceOption#COMPLETE}.
     */
    public static boolean isIncludeStackTraceComplete() {
        return getIncludeStackTraceOption() == IncludeStackTraceOption.COMPLETE;
    }

    /**
     * Fails unless {@code expected.equals(actual)}, {@code
     * actual.equals(expected)} and their hash codes are equal. This is useful
     * for testing the equals method itself.
     */
    public static void assertEqualsBothWays(Object expected, Object actual) {
        assertNotNull(expected);
        assertNotNull(actual);
        assertEquals("expected.equals(actual)", actual, expected);
        assertEquals("actual.equals(expected)", expected, actual);
        assertEquals("hashCode", expected.hashCode(), actual.hashCode());
    }

    /**
     * Fails unless {@code text} includes all {@code substrings}, in order.
     */
    public static void assertContains(String text, String... substrings) {

        int startingFrom = 0;
        for (String substring : substrings) {
            int index = text.indexOf(substring, startingFrom);
            assertTrue(String.format("Expected \"%s\" to contain substring \"%s\"", text, substring),
                    index >= startingFrom);
            startingFrom = index + substring.length();
        }

        String lastSubstring = substrings[substrings.length - 1];
        assertTrue(String.format("Expected \"%s\" to contain substring \"%s\" only once),",
                text, lastSubstring), text.indexOf(lastSubstring, startingFrom) == -1);
    }

    public static void awaitFullGc() {
        // GcFinalization *should* do it, but doesn't work well in practice...
        // so we put a second latch and wait for a ReferenceQueue to tell us.
        ReferenceQueue<Object> queue = new ReferenceQueue<Object>();
        WeakReference ref = new WeakReference<Object>(new Object(), queue);
        GcFinalization.awaitFullGc();
        try {
            assertSame("queue didn't return ref in time", ref, queue.remove(5000));
        } catch (IllegalArgumentException | InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    public static void awaitClear(WeakReference<?> ref) {
        // GcFinalization *should* do it, but doesn't work well in practice...
        // so we put a second latch and wait for a ReferenceQueue to tell us.
        Object data = ref.get();
        ReferenceQueue<Object> queue = null;
        WeakReference extraRef = null;
        if (data != null) {
            queue = new ReferenceQueue<Object>();
            extraRef = new WeakReference<Object>(data, queue);
            data = null;
        }
        GcFinalization.awaitClear(ref);
        if (queue != null) {
            try {
                assertSame("queue didn't return ref in time", extraRef, queue.remove(5000));
            } catch (IllegalArgumentException | InterruptedException e) {
                throw new RuntimeException(e);
            }
        }
    }
}
