package com.google.inject.spi;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Iterables;
import com.google.inject.AbstractModule;
import com.google.inject.Guice;
import com.google.inject.Inject;
import com.google.inject.Injector;
import com.google.inject.Key;
import com.google.inject.Provider;
import junit.framework.TestCase;

import java.util.Set;

public class HasDependenciesTest extends TestCase {

    /**
     * When an instance implements HasDependencies, the injected dependencies aren't used.
     */
    public void testInstanceWithDependencies() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            protected void configure() {
                bind(A.class).toInstance(new AWithDependencies());
            }
        });

        InstanceBinding<?> binding = (InstanceBinding<?>) injector.getBinding(A.class);
        assertEquals(ImmutableSet.<Dependency<?>>of(Dependency.get(Key.get(Integer.class))),
                binding.getDependencies());
    }

    public void testInstanceWithoutDependencies() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            protected void configure() {
                bind(A.class).toInstance(new A());
            }
        });

        InstanceBinding<?> binding = (InstanceBinding<?>) injector.getBinding(A.class);
        Dependency<?> onlyDependency = Iterables.getOnlyElement(binding.getDependencies());
        assertEquals(Key.get(String.class), onlyDependency.getKey());
    }

    public void testProvider() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            protected void configure() {
                bind(A.class).toProvider(new ProviderOfA());
            }
        });

        ProviderInstanceBinding<?> binding = (ProviderInstanceBinding<?>) injector.getBinding(A.class);
        Dependency<?> onlyDependency = Iterables.getOnlyElement(binding.getDependencies());
        assertEquals(Key.get(String.class), onlyDependency.getKey());
    }

    static class A {
        @Inject
        void injectUnusedDependencies(String unused) {
        }
    }

    static class ProviderOfA implements Provider<A> {
        @Inject
        void injectUnusedDependencies(String unused) {
        }

        public A get() {
            throw new UnsupportedOperationException();
        }
    }

    static class AWithDependencies extends A implements HasDependencies {
        public Set<Dependency<?>> getDependencies() {
            return ImmutableSet.<Dependency<?>>of(Dependency.get(Key.get(Integer.class)));
        }
    }

    static class ProviderOfAWithDependencies
            extends ProviderOfA implements ProviderWithDependencies<A> {
        public Set<Dependency<?>> getDependencies() {
            return ImmutableSet.<Dependency<?>>of(Dependency.get(Key.get(Integer.class)));
        }
    }
}
